"""Specific tasks."""
import copy
import json
import logging
import os
import shutil
import tempfile
import warnings
from collections import defaultdict
from inspect import cleandoc
from pathlib import Path

import sphinx.cmd.build  # pylint: disable=import-error
import sphinx.locale  # pylint: disable=import-error
from luigi_tools.util import get_dependency_graph

from data_validation_framework.rst_tools import RstFile

L = logging.getLogger(__name__)


_DEFAULT_REPORT_CONFIG = {
    "project": "Data validation report",
    "version": "1",
    "extensions": [
        "sphinx.ext.graphviz",
        "sphinx.ext.intersphinx",
        "sphinx.ext.napoleon",
        "sphinx.ext.todo",
        "rst2pdf.pdfbuilder",
    ],
    "author": "",
    "intersphinx_mapping": {
        "python": ("https://docs.python.org/3", None),
        "luigi": ("https://luigi.readthedocs.io/en/stable", None),
    },
}


def build_report_conf(dest, **kwargs):
    """Build the `conf.py` file use by Sphinx."""
    config = copy.deepcopy(_DEFAULT_REPORT_CONFIG)
    project = kwargs.pop("project", _DEFAULT_REPORT_CONFIG["project"])
    author = kwargs.pop("author", _DEFAULT_REPORT_CONFIG["author"])

    extensions = kwargs.pop("extensions", [])
    extensions.extend(config["extensions"])
    extensions = list(set(extensions))
    config["extensions"] = extensions

    config.update(
        {
            "pdf_documents": [
                ("index", "report", project, author),
            ],
            "latex_documents": [
                ("index", "report.tex", project, author, "howto"),
            ],
        }
    )

    config["intersphinx_mapping"].update(kwargs.pop("intersphinx_mapping", {}))

    config.update(kwargs)

    with open(dest, "w", encoding="utf-8") as config_file:
        config_file.writelines(["null = None\n"])
        config_file.writelines([f"{i[0]} = {json.dumps(i[1])} \n" for i in config.items()])


def build_subtree(node, deps, known_nodes=None):
    """Build the deps subtree of a given task."""
    if known_nodes is None:
        known_nodes = set()
    subtree = {}
    known_nodes.add(node)
    for t in deps[node]:
        if t not in known_nodes:
            subtree[t] = {}
            known_nodes.add(t)
    for t in subtree.keys():  # pylint: disable=consider-iterating-dictionary
        subtree[t] = build_subtree(t, deps, known_nodes)
    return subtree


def description_block(row, rst_file, indent=0, max_length=100):
    """Create a block for a failed feature."""
    list_indent = indent + 4
    rst_file.list([row.name], indent=indent)
    bullets = [
        f"return code: {row.ret_code}",
        f"comment: {row.comment}",
    ]
    if row.exception:
        bullets.append("exception:")
    rst_file.list(
        bullets,
        indent=list_indent,
        width=max_length - list_indent,
    )
    if row.exception:
        rst_file.code_block("Bash", indent + 8)
        exception_indent = indent + 12
        exception_width = max(20, max_length - exception_indent - 21)
        rst_file.exception(row.exception, exception_indent, width=exception_width)


def build_rst(rst_file, tree, level=1, use_data=True):
    """Build the RST file that can then be built with Sphinx."""
    for task_obj, subtree in tree.items():
        sublevel = level + 1
        rst_file.heading_level(level, task_obj.task_name)
        rst_file.add(cleandoc(task_obj.__specifications__))
        if use_data and task_obj.results is not None:
            df = task_obj.results.copy()
            succeeded = df.loc[df["is_valid"]]
            if len(succeeded) > 0:
                without_comment = succeeded.loc[
                    (succeeded["comment"].isnull()) | (succeeded["comment"] == "")
                ]
                if len(without_comment) > 0:
                    rst_file.heading_level(sublevel, "Validated features")
                    rst_file.add(", ".join(sorted(without_comment.index.astype(str).tolist())))

                with_comment = succeeded.loc[
                    (~succeeded["comment"].isnull()) & (succeeded["comment"] != "")
                ]
                if len(with_comment) > 0:
                    rst_file.heading_level(sublevel, "Validated features with warnings")
                    with_comment.apply(lambda x: description_block(x, rst_file), axis=1)
            if subtree:
                failed = df.loc[~df["is_valid"]]
                if len(failed) > 0:
                    rst_file.heading_level(sublevel, "Failed features")
                    rst_file.add(", ".join(sorted(failed.index.astype(str).tolist())))
            else:
                failed = df.loc[(~df["is_valid"]) & (~df["ret_code"].isnull())].sort_index()
                if len(failed) > 0:
                    rst_file.heading_level(sublevel, "Failed features")
                    failed.apply(lambda x: description_block(x, rst_file), axis=1)

        if subtree:
            rst_file.heading_level(sublevel, "Sub tasks")
            build_rst(rst_file, subtree, sublevel + 1)

        rst_file.newline()


# pylint: disable=too-many-statements
def make_report(root, use_data=True, config=None):
    """Create the report based on the current workflow."""
    g = get_dependency_graph(root)

    deps = defaultdict(list)
    for parent, child in g:
        deps[parent].append(child)

    tree = {root: build_subtree(root, deps)}

    # Build the PDF in a temporary directory then move the file
    with tempfile.TemporaryDirectory() as tmpdirname:
        path = Path(tmpdirname)
        source_path = path / "source_report"
        build_path = path / "build_report"
        source_path.mkdir(parents=True, exist_ok=True)
        build_report_conf(source_path / "conf.py", **(config or {}))

        # Set environment variable to make latexmk quiet (only used when report_type == latexpdf)
        os.environ["LATEXMKOPTS"] = "-quiet -f"

        # Build the RST buffer
        rst_file = RstFile(source_path / "index.rst")
        rst_file.title("Validation report")
        rst_file.toc_tree(maxdepth=5, numbered=True)

        build_rst(rst_file, tree, use_data=use_data)

        # Write the RST file
        rst_file.write()

        # Build the PDF file from the RST file
        # Run the following Sphinx command:
        # sphinx-build -M latexpdf <source_report> <build_report> -W
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            sphinx.cmd.build.make_main(
                ["-M", root.report_type, str(source_path), str(build_path), "-q"]
            )

        # Copy the report to its final destination
        report_path = root.report_path or f"report_{root.task_name}.pdf"
        if root.result_path is not None:  # pragma: no cover
            report_path = Path(root.result_path) / report_path

        if root.report_type == "pdf":
            input_path = build_path / "pdf" / "report.pdf"
        elif root.report_type == "latexpdf":  # pragma: no cover
            input_path = build_path / "latex" / "report.pdf"
        else:  # pragma: no cover
            input_path = build_path / root.report_type / "report.pdf"

        Path(report_path).parent.mkdir(parents=True, exist_ok=True)
        shutil.copy(input_path, report_path)
        L.info("The report was generated here: %s", report_path)



"""Specific targets."""
import warnings

from luigi_tools.target import OutputLocalTarget


class TaggedOutputLocalTarget(OutputLocalTarget):
    """Target with tagged output path."""

    _already_changed = False

    @classmethod
    def set_default_prefix(cls, prefix, quiet=False):
        """Set the default prefix of the class."""
        if cls._already_changed and not quiet and str(prefix) != str(cls.get_default_prefix()):
            warnings.warn(
                f"The default prefix of '{cls.__name__}' was already changed before, changing it "
                f"again might lead to unexpected behavior (old value: {cls.get_default_prefix()} ; "
                f"new value: {prefix}."
            )
        cls._already_changed = True
        super().set_default_prefix(prefix)


class _WithTaskNameMixin:
    """Mixin to add the task name as attribute."""

    def __init__(self, *args, task_name=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.task_name = task_name


class ReportTarget(_WithTaskNameMixin, TaggedOutputLocalTarget):
    """Specific target for BaseValidationTask reports."""


class DataDirectoryTarget(_WithTaskNameMixin, TaggedOutputLocalTarget):
    """Specific target for BaseValidationTask reports."""



"""Some tools to generate RST files."""
import re
import textwrap


class RstFile:
    """Class to help creating proper RST files."""

    def __init__(self, file_path, mode="w"):
        if not file_path:
            raise ValueError("The 'file_path' argument must be a not empty string.")
        self.file_path = file_path
        self.mode = mode
        self._data = []
        pattern = r"(\x9B|\x1B\[)[0-?]*[ -\/]*[@-~]"
        self._ANSI_re_compiled = re.compile(pattern)

    def add(self, content, indent=0):
        """Add a single line to the internal buffer."""
        if isinstance(content, list):
            self._data.extend(self._indent(content, indent=indent))
        else:
            self._data.append(self._indent(content, indent=indent))

    def newline(self, count=1):
        """Add an empty line to the internal buffer."""
        if not isinstance(count, int) or count <= 0:
            raise ValueError("Count of newlines must be a positive int.")

        self.add(["\n"] * count)

    def write(self):
        """Write the internal buffer into the given file."""
        with open(self.file_path, self.mode, encoding="utf-8") as f:
            f.write("\n".join(self._data))
            f.write("\n")

    def get_buffer(self):
        """Get the internal buffer."""
        return self._data

    def clear(self):
        """Clear the internal buffer."""
        self._data = []

    @staticmethod
    def bold(string):
        """Bold the given string."""
        return f"**{string}**"

    @staticmethod
    def emph(string):
        """Emph the given string."""
        return f"*{string}*"

    @staticmethod
    def pre(string):
        """Pre the given string."""
        return f"``{string}``"

    @staticmethod
    def inline_link(text, link):
        """Create a link from the given string and URL."""
        return f"`{text} <{link}>`_"

    @staticmethod
    def _indent(content, indent=0):
        """Indent the given string."""
        if indent == 0:
            return content
        indent = " " * indent
        if isinstance(content, list):
            return ["".join([indent, line]) for line in content]
        return "".join([indent, content])

    def title(self, text, indent=0, char="="):
        """Add a title to the internal buffer."""
        line = char * len(text)
        self.add([line, text, line], indent=indent)
        self.newline()

    def toc_tree(self, toc_type="toctree", maxdepth=None, numbered=True, indent=0):
        """Add a toc tree to the internal buffer."""
        self.newline()

        if toc_type == "toctree":
            self.add(".. toctree::", indent=indent)

            if maxdepth:
                self.add(f"   :maxdepth: {maxdepth}", indent=indent)

            if numbered:
                self.add("   :numbered:", indent=indent)

        elif toc_type == "contents":
            self.add(".. contents::", indent=indent)

        else:
            raise ValueError("Unknown TOC tree type")

        self.newline()

    def heading(self, text, char, indent=0):
        """Add a heading to the internal buffer."""
        self.newline()
        self.add([text, char * len(text)], indent=indent)
        self.newline()

    def h1(self, text, indent=0):
        """Add a heading of level 1 to the internal buffer."""
        self.heading(text, char="=", indent=indent)

    def h2(self, text, indent=0):
        """Add a heading of level 2 to the internal buffer."""
        self.heading(text, char="-", indent=indent)

    def h3(self, text, indent=0):
        """Add a heading of level 3 to the internal buffer."""
        self.heading(text, char="~", indent=indent)

    def h4(self, text, indent=0):
        """Add a heading of level 4 to the internal buffer."""
        self.heading(text, char="+", indent=indent)

    def h5(self, text, indent=0):
        """Add a heading of level 5 to the internal buffer."""
        self.heading(text, char="^", indent=indent)

    def h6(self, text, indent=0):
        """Add a heading of level 6 to the internal buffer."""
        self.heading(text, char=";", indent=indent)

    def h7(self, text, indent=0):
        """Add a heading of level 7 to the internal buffer."""
        self.heading(text, char="_", indent=indent)

    def h8(self, text, indent=0):
        """Add a heading of level 8 to the internal buffer."""
        self.heading(text, char="*", indent=indent)

    def h9(self, text, indent=0):
        """Add a heading of level 9 to the internal buffer."""
        self.heading(text, char="#", indent=indent)

    def h10(self, text, indent=0):
        """Add a heading of level 10 to the internal buffer."""
        self.heading(text, char=">", indent=indent)

    def h11(self, text, indent=0):
        """Add a heading of level 11 to the internal buffer."""
        self.heading(text, char="<", indent=indent)

    def h12(self, text, indent=0):
        """Add a heading of level 12 to the internal buffer."""
        self.heading(text, char=",", indent=indent)

    def heading_level(self, level, *args, **kwargs):
        """Add a heading of given level to the internal buffer."""
        level_headings = {
            0: self.title,
            1: self.h1,
            2: self.h2,
            3: self.h3,
            4: self.h4,
            5: self.h5,
            6: self.h6,
            7: self.h7,
            8: self.h8,
            9: self.h9,
            10: self.h10,
            11: self.h11,
            12: self.h12,
        }

        if level not in level_headings:  # pragma: no cover
            raise ValueError(
                f"The 'level' value must be in {list(level_headings.keys())} but is {level}."
            )

        return level_headings[level](*args, **kwargs)

    def code_block(self, language="", indent=0):
        """Add a code-block directive to the internal buffer."""
        self.newline()
        self.add([f".. code-block:: {language}"], indent)
        self.newline()

    def exception(self, exception=None, indent=0, width=None):
        """Format and add an exception to the internal buffer."""
        self.newline()
        if exception is None:
            self.add("")
        else:
            # Clean ANSI escape sequences (console colors)
            exception = [self._ANSI_re_compiled.sub("", i) for i in exception.split("\n")]

            if width is not None:
                exception = [
                    j
                    for i in exception
                    for j in textwrap.fill(
                        i,
                        width=width,
                        break_on_hyphens=False,
                        break_long_words=True,
                    ).split("\n")
                ]
            self.add(exception, indent)
        self.newline()

    def list(self, elements, width=None, indent=0, bullet="*"):
        """Add list elements to the internal buffer."""
        for elem in elements:
            elem = f"{bullet} {elem}"
            if width is not None:
                elem = textwrap.fill(
                    elem,
                    width=width,
                    break_on_hyphens=False,
                    break_long_words=True,
                    expand_tabs=False,
                    initial_indent=" " * indent,
                    replace_whitespace=False,
                    subsequent_indent=" " * (indent + 2),
                )
            self.add(elem)
